package jartege ; 

import java.lang.reflect.* ; 
import java.io.* ;
import java.util.* ; 
import org.jmlspecs.jmlrac.runtime.* ;

/**
 * The default object generator associated with a class.
 */

class DefaultObjectGenerator extends ObjectGenerator {

   public DefaultObjectGenerator() {
      super() ; 
   }

   public DefaultObjectGenerator(Classe classe) {
      super() ; 
      setClasse(classe) ; 
   }

   /**
    * Generates one object of the class associated with this object 
    * generator.
    * @param nonNull if nonNull = false, then the method may generate 
    *                a null object.
    * @param file the generated file
    */

   Expression generateAnObject(boolean nonNull, FileWriter file)
      throws IOException {

      Trace.afficher(4, "generateAnObject(" + nonNull + ", ...) ->") ;

      Classe laClasse = getClasse() ; 

      Expression resultat = null ;
      int nb = laClasse.lesObjets.size() ;
      if (nonNull && nb != 0) {
         nb = nb - 1 ;
      }

      // Seuil de cration d'un nouvel objet
      double probCreer = getCreationProbability().theFunction(nb) ;
      double val = RandomValue.floatValue(1) ;

      boolean trouve = false ;

      if (val < probCreer) {

         if (laClasse.estPrimitif()) {
            resultat = laClasse.genererValeurTypeBase() ;

         } else if (laClasse.theClass.isArray()) {
            // Construction d'un tableau
            int taille = RandomValue.intValue(9) ;
            String nomVar = GenerateurNom.nouveau(laClasse.theClass) ;
            Class componentType = laClasse.theClass.getComponentType() ;
            Classe composant = laClasse.leDictionnaire.getClasse(componentType) ;
            Object valeur = Array.newInstance(componentType, taille) ;
            Expression[] tabValeur = new Expression[taille] ;

            for (int i = 0 ; i < taille ; i++) {
               tabValeur[i] = composant.genererUnObjet(false, file) ;
               Array.set(valeur, i, tabValeur[i].getValeur()) ;
            }

            Chaine.ecrire(file,
               "      " + laClasse.nom + " " + nomVar + " = {") ;

            for (int i = 0 ; i < taille ; i++) {
               Chaine.ecrire(file, tabValeur[i].toString()) ;
               if (i != taille - 1) {
                  Chaine.ecrire(file, ", ") ;
               }
            }
            Valeur res = new Valeur(nomVar, laClasse, valeur) ;
            laClasse.ajouter(res) ; 
            resultat = res ; 

            Chaine.ecrire(file, "} ; \n") ;

         } else {

            List lesCons = new Vector(laClasse.lesConstructeurs) ;
            Operation cons ;

            while (!trouve && lesCons.size() != 0) {
               cons = (Operation) (ElementAPoids.choisir(lesCons)) ;
               resultat = cons.genererUnAppel(file) ;
               if (resultat != null &&
                   !(resultat.valeur instanceof JMLEntryPreconditionError)) {
                  trouve = true ;
               } else {
                  lesCons.remove(cons) ;
               }
            }
         }
      } 

      if (resultat == null) { 
         resultat = laClasse.choisirValeur(nonNull) ;
      }

      if (resultat == null) {
         if (nonNull) {
            Trace.afficher(5, "generateAnObject(true, ...) retourne null !!!!") ; 
         } else {
            Trace.afficher(5, "generateAnObject(false, ...) retourne null !") ; 
         }
      } 

      return resultat ;
   }

}

