package jartege ; 

import java.io.* ;

/**
 * Class of objects generators.
 * An object generator associated with a class allows one to generate objects
 * of this class with various algorithms, which are described in 
 * subclasses of ObjectGenerator. 
 */

abstract class ObjectGenerator {

   // The class associated with this object generator. 
   private Classe theClass ; 

   // The creation probability function associated with this object generator.
   private CreationProbability theCreationProbability ; 

   /**
    * Constructor.
    */

   public ObjectGenerator(Classe theClass) {
      this() ; 
      this.theClass = theClass ; 
   }

   /**
    * Constructor.
    */

   public ObjectGenerator() {
      this.theCreationProbability = ClassTester.creationProbabilityNormal ;
   }


   /**
    * The class associated with this object generator.
    */
  
   Classe getClasse() {
      return theClass ; 
   }

   /**
    * Sets the class associated with this object generator.
    */

   void setClasse(Classe theClass) {
      this.theClass = theClass ; 
   }

   /**
    * The creation probability function associated with this object generator.
    */
 
   CreationProbability getCreationProbability() {
      return theCreationProbability ;
   }

   /**
    * Sets the creation probability function associated with this object 
    * generator.
    */

   void setCreationProbability(CreationProbability theCreationProbability) {
      this.theCreationProbability = theCreationProbability ;
   }

   /**
    * Generate one object of the class associated with this object 
    * generator.
    * @param nonNull if nonNull = false, then the method may generate 
    *                a null object.
    * @param file the generated file
    */

   abstract Expression generateAnObject(boolean nonNull, FileWriter file)
      throws IOException ; 

}

