package jartege ; 

import java.lang.reflect.* ; 
import java.util.* ; 
import java.io.* ; 

import org.jmlspecs.jmlrac.runtime.JMLAssertionError ; 

/**
 * Classe des oprations.
 */

abstract class Operation extends ElementAPoids {

   // La classe qui contient cette opration
   Classe laClasse ;

   // Vecteur des methodes qui permettent de gnrer les valeurs
   // des paramtres
   List methodesGenerantParametres ;

   // Number of calls of this operation
   int numberOfCalls ; 

   // The storeResult flag of this operation indicates whether or not
   // results of this operation must be stored in objects
   private boolean storeResult ; 

   /**
    * Constructeur.
    */
   Operation(Classe laClasse) {
      this.laClasse = laClasse ; 
      this.storeResult = true ; 
      methodesGenerantParametres = new Vector() ;
      numberOfCalls = 0 ; 
   }

   /**
    * Sets the storeResult flag to the specified boolean value
    */
   void setStoreResult(boolean storeResult) {
      this.storeResult = storeResult ; 
   }

   /**
    * Returns the storeResult flag value.
    */
   boolean storeResult() {
      return storeResult ; 
   }

   /**
    * Calcule le vecteur methodesGenerantParametres.
    */
   void calculeVecteurMGP() {

      String nomBaseOp = nomBaseMGP() ; 

      Method[] tabMethodes ; 

      if (laClasse.classeMGP != null) {
         tabMethodes = laClasse.classeMGP.theClass.getMethods() ; 
      } else {
         tabMethodes = new Method[0] ; 
      }

      Class[] typesParametres = typeParametres() ; 

      for (int i = 0 ; i < typesParametres.length ; i++) {
         boolean trouve = false ; 
         int j ; 
         int num = i + 1 ; 
         String nomOp = nomMGP(nomBaseOp, num) ; 
         for (j = 0 ; !trouve && j < tabMethodes.length ; j++) {
            trouve = tabMethodes[j].getName().equals(nomMGP(nomBaseOp, num)) ; 
         }
         if (trouve) {
            methodesGenerantParametres.add(tabMethodes[j - 1]) ; 
         } else {
            methodesGenerantParametres.add(null) ; 
         }
      }
   }

   /**
    * Retourne vrai si et seulement si l'opration est une opration
    * de classe.
    */
   abstract boolean estStatique() ; 

   /**
    * Retourne vrai si et seulement si l'opration est un constructeur.
    */
   boolean estConstructeur() {
      return false ; 
   }

   /**
    * Retourne e type des paramtres de l'opration.
    */
   abstract Class[] typeParametres() ; 

   /**
    * Calcule le nom de base de la mthode gnrant les paramtres.
    */
   String nomBaseMGP() {
      Class[] typesParametres = typeParametres() ; 
      String nomOp = "JRT_" + getNom() + "_" ; 
 
      for (int i = 0 ; i < typesParametres.length ; i++) {
         Class theParam = typesParametres[i] ; 
         String nomParam = laClasse.leDictionnaire.getClasse(theParam).getNom() ; 
         String nomType = nomParam.replace('[', '_').replace(']', '_') ;  
         nomOp = nomOp + nomType + "_" ;
      }
      return nomOp ; 
   }

   /**
    * Calcule le nom de la mtode gnrant le parametre numro i.
    */
   String nomMGP(String nomBase, int i) {
      return nomBase + i ; 
   }

   /**
    * Evaluates a call of this operation.
    */
   abstract Object eval(Appel appel) throws 
      InstantiationException, IllegalAccessException,
      IllegalArgumentException, InvocationTargetException ; 

   /**
    * A call as a String.
    */
   abstract String toString(Appel appel) ; 

   /**
    * The return type of this operation.
    */
   abstract Class typeRetour() ; 

   /**
    * The declared exceptions of this operation.
    */
   Class[] typesExceptions() {
      return new Class[0] ; 
   }

   /**
    * Teste si la classe c2 est la classe des objets correspondant 
    *  la classe primitive c1.
    */
   static boolean presqueEgal(Classe c1, Classe c2) {
      if (c1.theClass == Integer.TYPE) {
         return (c2.getNom().equals("Integer")) ; 
      } else if (c1.theClass == Long.TYPE) {
         return (c2.getNom().equals("Long")) ; 
      } else if (c1.theClass == Float.TYPE) {
         return (c2.getNom().equals("Float")) ; 
      } else if (c1.theClass == Double.TYPE) {
         return (c2.getNom().equals("Double")) ; 
      } else if (c1.theClass == Byte.TYPE) {
         return (c2.getNom().equals("Byte")) ; 
      } else if (c1.theClass == Short.TYPE) {
         return (c2.getNom().equals("Short")) ; 
      } else if (c1.theClass == Boolean.TYPE) {
         return (c2.getNom().equals("Boolean")) ; 
      } else if (c1.theClass == Character.TYPE) {
         return (c2.getNom().equals("Character")) ; 
      } else {
         return c1 == c2 ; 
      }
   }
   

   /**
    * Array decomposition.
    */
   void tableaux(String prefixe, Class staticType, Object resAppel) 
   throws IOException {

      Trace.afficher(6, "tableaux -> " + prefixe + ", " + resAppel) ;

      String nom ; 
      Class dynType = resAppel.getClass() ; 

      if (dynType.isArray()) {
         Class component = dynType.getComponentType() ; 
         Classe composant = laClasse.leDictionnaire.getClasse(component) ; 
         for (int i = 0 ; i < Array.getLength(resAppel) ; i++) {
            Object object = Array.get(resAppel, i) ; 
            if (object != null) {
               Class cl = object.getClass() ; 
               Classe c = laClasse.leDictionnaire.getClasse(cl) ; 
               Valeur var ; 
               if (presqueEgal(composant, c)) {
                  nom = prefixe + "[" + i + "]" ; 
                  var = new Valeur(nom, composant, object) ; 
                  composant.ajouter(var) ; 
               } else {
                  nom = "((" + c.getNomComplet() + ") " + prefixe + 
                        "[" + i + "])" ; 
                  var = new Valeur(nom, c, object) ; 
                  c.ajouter(var) ; 
               }
               tableaux(nom, component, object) ; 
            }
         }

      }
      Trace.afficher(6, "tableaux <-") ; 
   }

   /**
    * Collections decomposition.
    */
   void collection(String nom, Object resAppel, FileWriter fichier, int i) 
         throws IOException {
      if (resAppel instanceof Collection) {
         Collection coll = (Collection) resAppel ;
         int taille = coll.size() ;
         if (taille != 0) {
            Object[] tab = coll.toArray() ; 
            String nomTab = GenerateurNom.nouveau(tab.getClass()) ; 
            Chaine.ecrire(fichier,
                "      Object[] " + nomTab + " = " + nom + 
                ".toArray() ; \n") ; 
            tableaux(nomTab, tab.getClass(), tab) ; 
         }
      }
   }

   // Choix d'un parametre
   Valeur choisirParametrePrimitif(int numero, Valeur obj, Classe typeParam) {
      Trace.afficher(5, "choisirParametrePrimitif -> " + numero) ; 
      Valeur resultat = null ;
      Method methGen = (Method) (methodesGenerantParametres.get(numero)) ;
      if (methGen == null) {
         // Pas de methode de generation de parametre
         resultat = typeParam.genererValeurTypeBase() ;
      } else {
         // Une mthode est disponible pour gnrer un parametre
         resultat = genererParametreMGP(methGen, obj) ;
      }
      return resultat ;
   }


   // Gnration d'un parametre
   Expression genererParametre
             (int numero, Expression obj, Classe typeParam, FileWriter fichier) 
          throws IOException {

      Expression resultat = null ; 
      Method methGen = (Method) (methodesGenerantParametres.get(numero)) ; 
      if (methGen == null) {
         // Pas de methode de generation de parametre
         resultat = typeParam.genererUnObjet(false, fichier) ;
      } else {
         // Une mthode est disponible pour gnrer un parametre
         resultat = genererParametreMGP(methGen, obj) ; 
      }
      return resultat ; 
   }
       

   /**
    * Generation d'une valeur de parametre  l'aide d'une methode
    * de gnration d'un paramtre.
    */

   Valeur genererParametreMGP (Method methGen, Expression obj) {
      Trace.afficher(5, "genererParametreMGP -> ") ; 

      Valeur resultat = null ;

      String nomClasseMGP = laClasse.classeMGP.getNomComplet() ; 

      Object[] param = new Object[0] ; 
      Object o = null ; 

      if (obj == null && !Modifier.isStatic(methGen.getModifiers())) {
         System.err.println("Erreur dans la classe " + nomClasseMGP + ".") ; 
         System.err.println(methGen.getName() + " devrait etre statique.") ; 
         System.exit(1) ; 
      }

      Object object ; 
      if (obj == null) {
         object = null ; 
      } else {
         if (obj.objetMGP == null) {
            try {
               Class clas = laClasse.classeMGP.theClass ; 
               Constructor c = clas.getConstructors()[0] ; 
               Object[] paramCons = {obj.valeur} ; 
               Object objMGP = c.newInstance(paramCons) ; 
               obj.objetMGP = objMGP ; 
            } catch (Exception except) {
               System.err.println("Erreur dans la classe " + 
                  nomClasseMGP) ; 
               System.err.println("Unique constructeur") ; 
               System.err.println("   public " + nomClasseMGP +
                  "(" + laClasse.getNom() + " object) {...} \n" + 
                  "attendu.") ; 
               System.err.println(obj.toString()) ; 
               except.printStackTrace() ;
               System.exit(1) ; 
            }
         }
         object = obj.objetMGP ; 
      }

      Class returnType = methGen.getReturnType() ;
      Classe classe = laClasse.leDictionnaire.getClasse(returnType) ;

      try {
         Trace.afficher(5, "invocation") ; 
         o = methGen.invoke(object, param) ; 
         Valeur val = new Valeur(classe, o) ; 
         resultat = val ; 
      } catch (InvocationTargetException except) {
         Throwable target = except.getTargetException() ; 
         if (!(target instanceof JMLAssertionError)) {
            System.err.println("Warning : l'execution d'un appel de " +
               methGen.getName() + " a leve l'exception " + target) ;
         }
         resultat = classe.genererValeurTypeBase() ;
      } catch (IllegalAccessException except) {
         System.err.println("Erreur l'execution d'un appel de " +
            methGen.getName() + ".") ;
         System.err.println("Operation inaccessible.") ;
         System.err.println("  - La classe " + nomClasseMGP +
            " n'est pas accessible (devrait etre publique)") ;
         System.err.println("  - L'operation " +
            methGen.getName() +
            " n'est pas accessible (devrait etre publique)") ;
          System.exit(1) ; 
      } catch (IllegalArgumentException except) {
         System.err.println("Erreur l'execution d'un appel de " +
            methGen.getName() + ".") ;
         System.err.println("Parametre incorrect.") ; 
         System.err.println(
            "Les fichiers .class ne sont peut-etre pas a jour.") ;
         System.exit(1) ; 
      }
      return resultat ; 
   }

   Valeur genererAppel(FileWriter fichier, Appel appel, Object resAppel) 
   throws IOException {

      Trace.afficher(5, "genererAppel -> appel = " + appel + 
         ", resAppel = " + resAppel) ; 
      Valeur resultat ; 

      Class returnType = typeRetour() ; 
      Classe typeRetour = laClasse.leDictionnaire.getClasse(returnType) ;  

      boolean estVoid = (returnType == Void.TYPE) ; 
      boolean tryOuvert = false ; 
      boolean erreurAssertion = false ; 
      ResultatException resExcept = null ; 
      String prefixe = "         " ; 
      String prefixe2 = prefixe ; 

      String nomVar = "" ; 
      Valeur laValeur = null ; 

      numberOfCalls++ ; 

      if (! (this instanceof Constructeur)) {
         // Chaine.ecrire(fichier, prefixe + "_nb_calls++ ; \n") ; 
      }

      if (resAppel instanceof ResultatException) {
         resExcept = (ResultatException) resAppel ; 
         if (Constantes.FOR_JUNIT && 
             resExcept.theExcept instanceof JMLAssertionError) {
            erreurAssertion = true ; 
         }
         // tryOuvert = true ; 
      }

      if (!estVoid && storeResult()) {
         nomVar = GenerateurNom.nouveau(returnType) ;
         if (tryOuvert) {
            laValeur = laClasse.leDictionnaire.uneValeur(typeRetour) ; 
            Chaine.ecrire(fichier,
               prefixe + typeRetour.getNomComplet() + 
               " " + nomVar +
               " = " + laValeur.toString() + " ; \n") ; 
         } else {
            if (resAppel instanceof ResultatException) {
               laValeur = laClasse.leDictionnaire.uneValeur(typeRetour) ;
            } else {
               laValeur = new Valeur(nomVar, typeRetour, resAppel) ; 
               typeRetour.ajouter(laValeur) ; 
            }
         }
         resultat = laValeur ; 
      } else {
         resultat = laClasse.leDictionnaire.varVoid() ; 
      }

      if (tryOuvert) {
         Chaine.ecrire(fichier, prefixe + "try { \n") ;
         prefixe2 = prefixe + "   " ; 
      } 

      Chaine.ecrire(fichier, prefixe2) ; 

      if (!estVoid && storeResult()) {
         if (!tryOuvert) {
            Chaine.ecrire(fichier, typeRetour.getNomComplet() + " ") ; 
         }
         Chaine.ecrire(fichier, nomVar + " = ") ; 
      }

      Chaine.ecrire(fichier, toString(appel) + " ; \n") ; 

      if (tryOuvert) {
         if (Constantes.FOR_JUNIT && erreurAssertion) {
            Chaine.ecrire(fichier, 
               prefixe + "} catch (JMLAssertionError except) { \n" +
               prefixe2 + "_message = _message + \"\\nAssertion error: \"" + 
               " + except.getMessage() ; \n" + 
               prefixe2 + "_failure = true ; \n") ; 
         } else {
            Chaine.ecrire(fichier, 
               prefixe + "} catch (Throwable except) { \n" + 
               prefixe2 + "System.out.print(\"Exception raised and caught" + 
               " in test file: \") ; \n" +
               prefixe2 + "except.printStackTrace() ; \n" + 
               prefixe2 + "System.out.println() ; \n") ;  
         }
         Chaine.ecrire(fichier, prefixe + "} \n") ; 
      }

      if (!estVoid && resAppel != null && !tryOuvert) {
         Class c = resAppel.getClass() ; 
         Classe cl = laClasse.leDictionnaire.getClasse(c) ; 

         // Generation d'un cast pour une sous classe
         if (returnType.isAssignableFrom(c) && 
            (cl != typeRetour)) { 

            nomVar = "((" + cl.getNomComplet() + ") " + nomVar + ")" ; 
            laValeur = new Valeur(nomVar, cl, resAppel) ; 
            cl.ajouter(laValeur) ; 
         }

         // Dcomposition des collections
         collection(nomVar, resAppel, fichier, 1) ; 

         // Dcomposition des tableaux
         tableaux(nomVar, returnType, resAppel) ; 
      }

      Trace.afficher(5, "genererAppel <- resultat = " + resultat) ; 
      return resultat ; 
   }


   /**
    * Gnre un appel de cette opration
    * Retourne null si on n'a pas reussi a generer un appel 
    * parce qu'on n'a pas reussi soit a generer le parametre implicite, 
    * soit a generer un appel qui respecte la precondition
    */

   Expression genererUnAppel(FileWriter fichier) throws IOException {

      Trace.afficher(4, "genererUnAppel -> " + nom) ; 

      Expression paramImpl = null ; 
      Expression[] tabParam ; 
      Class[] typeParametres = typeParametres() ;  
      Expression resultat = null ; 
 
      boolean ok = true ; 

      if (!estStatique() && !estConstructeur()) {
         paramImpl = laClasse.genererUnObjet(true, fichier) ;
         if (paramImpl == null) {
            ok = false ; 
         } else {
         // Trace.afficher(5, "paramImpl = " + paramImpl.getNomComplet()) ; 
         }
      }
      if (ok) {
         tabParam = new Valeur[typeParametres.length] ; 
         for (int i = 0 ; i < typeParametres.length ; i++) {
            Classe typeParam = 
               laClasse.leDictionnaire.getClasse(typeParametres[i]) ;
            tabParam[i] = genererParametre(i, paramImpl, typeParam, fichier) ; 
         }

         Appel appel = new Appel(this, paramImpl, tabParam) ; 
         Object resAppel = appel.eval() ; 
         if (!ResultatException.precondErreur(resAppel)) {
            resultat = genererAppel(fichier, appel, resAppel) ;
         }
      }
      if (resultat == null) {
         resultat = genererUnAppelBis(fichier) ; 
      }

      if (resultat == null) {
         Trace.afficher(1, "echec de generation d'appel pour la methode : " + 
                           nom) ; 
      }
      return resultat ;
   }

   /**
    * Essaie de generer un appel d'opration.
    * Une premire tentative ayant choue, on essaie de gnrer
    * un appel de facon plus systmatique.
    */

   Valeur genererUnAppelBis(FileWriter fichier) throws IOException {

      Trace.afficher(4, "genererUnAppelBis -> " + nom) ; 

      // Le rsultat
      Valeur resultat = null ;

      // Les possibilits pour les paramtres explicites
      List[] tabParam ;
      // Le type des paramtres
      Class[] typeParametres = typeParametres() ;

      if (!estStatique() && !estConstructeur()) {

         boolean fini = false ;
         int pi = 0 ;
         List lesObjets = laClasse.lesObjets ;
         while (pi < lesObjets.size() && !fini) {
            Valeur paramImpl = (Valeur) lesObjets.get(pi) ; 
            if (paramImpl.getValeur() != null) {
               tabParam = construireListeParametresExplicites(paramImpl) ; 
               resultat = chercherAppelCorrect(paramImpl, tabParam, fichier) ; 
               fini = (resultat != null) ; 
            }
            pi++ ; 
         }
      } else {
         tabParam = construireListeParametresExplicites(null) ;
         resultat = chercherAppelCorrect(null, tabParam, fichier) ;
      }
      return resultat ; 
   }

   /**
    * Execute les appels avec les differents parametres jusqu'a trouver
    * un appel correct (respectant la precondition).
    */
   Valeur chercherAppelCorrect
         (Valeur paramImpl, List[] tabParam, FileWriter fichier) 
         throws IOException {
      Valeur resultat = null ; 
      boolean fini = false ; 
      Class[] typeParametres = typeParametres() ; 
      int[] max = new int[typeParametres.length] ;
      for (int i = 0 ; i < typeParametres.length ; i++) {
         max[i] = tabParam[i].size() ;
      }
      IndexIterator it = new IndexIterator(max) ;
      while (it.hasNext() && !fini) {
         int[] ind = it.next() ;
         Valeur[] objetParam = new Valeur[ind.length] ;
         for (int i = 0 ; i < ind.length ; i++) {
            objetParam[i] = (Valeur) tabParam[i].get(ind[i]) ;
         }
         Appel appel = new Appel(this, paramImpl, objetParam) ;
         Object resAppel = appel.eval() ;
         if (!ResultatException.precondErreur(resAppel)) {
            resultat = genererAppel(fichier, appel, resAppel) ;
            fini = true ;
         }
      }
      return resultat ; 
   }

   // Construction de la liste des possibilites pour les parametres 
   // explicites
   List[] construireListeParametresExplicites(Valeur paramImpl) {
      // Le rsultat
      List[] tabParam ; 
      Class[] typeParametres = typeParametres() ; 
      tabParam = new Vector[typeParametres.length] ;
      for (int i = 0 ; i < typeParametres.length ; i++) {
         tabParam[i] = new Vector() ; 
         Classe typeParam =
            laClasse.leDictionnaire.getClasse(typeParametres[i]) ;
         if (typeParam.estPrimitif()) {
            for (int j = 0 ; j < 20 ; j++) {
               tabParam[i].add(
                  choisirParametrePrimitif(i, paramImpl, typeParam)) ; 
            }
         } else {
            List lesObjets = typeParam.lesObjets ;
            if (lesObjets.size() == 0) {
               System.out.println(
                  "Construction d'une liste de parametres explicites de type " +
                  typeParam.getNomComplet() + " : ") ;
               System.out.println("La liste d'objets est vide") ;
            }
            for (int j = 0 ; j < lesObjets.size() ; j++) {
               Valeur objet = (Valeur) lesObjets.get(j) ;
               tabParam[i].add(objet) ; 
            } 
         }
      }
      return tabParam ; 
   }

   /**
    * String corresponding to this operation.
    */
   public String toString() {
      String result = typeRetour().getName() + " " + getNom() + "(" ; 
      Class[] params = typeParametres() ; 
      for (int i = 0 ; i < params.length ; i++) {
         result = result + params[i].getName() ; 
         if (i != params.length - 1) {
            result = result + ", " ; 
         }
      }
      result = result + ")" ; 
      return result ; 
   }
}

