package jartege ; 

import java.util.* ; 
import java.io.* ; 

/**
 * Classe des valeurs.
 */

class Valeur extends Expression {

   // Representation de cette valeur sous forme de chaine de 
   // caractres ; 
   String nom ; 

   /**
    * Constructeur de valeur.
    */
   Valeur(String nom, Classe laClasse, Object valeur) {
      super() ; 
      this.nom = nom ; 
      this.laClasse = laClasse ; 
      this.valeur = valeur ; 
      objetMGP = null ; 
   }

   /**
    * Constructeur de valeur d'un type primitif.
    */
   Valeur(Classe laClasse, Object valeur) {
      super() ; 
      this.nom = nom ; 
      this.laClasse = laClasse ; 
      this.valeur = valeur ; 
      objetMGP = null ; 
      // laClasse.ajouter(this) ; 
      String nomClasse = laClasse.getNom() ; 
      if (valeur == null) {
         nom = "(" + nomClasse + ") null" ; 
      } else if (nomClasse.equals("int")) {
         nom = ((Integer) valeur).intValue() + "" ; 
      } else if (nomClasse.equals("long")) {
         nom = ((Long) valeur).longValue() + "L" ; 
      } else if (nomClasse.equals("float")) {
         nom = ((Float) valeur).floatValue() + "F" ;  
      } else if (nomClasse.equals("boolean")) {
         nom = ((Boolean) valeur).booleanValue() + "" ;  
      } else if (nomClasse.equals("double")) {
         nom = ((Double) valeur).doubleValue() + "" ;  
      } else if (nomClasse.equals("char")) {
         nom = charToString(((Character) valeur).charValue()) ; 
      } else if (nomClasse.equals("byte")) {
         nom = "(byte) " + ((Byte) valeur).byteValue() ;  
      } else if (nomClasse.equals("short")) {
         nom = "(short) " + ((Short) valeur).shortValue() ;  
      } else if (nomClasse.equals("String")) {
         String s = (String) valeur ; 
         nom = stringToString(s) ; 
      } else {
         System.out.println("Erreur de constructeur de Valeur.") ; 
         System.exit(1) ; 
      }
   }

   /**
    * Convertit un entier n compris entre 0 et 15 en un caractre
    * compris entre '0' et '9' ou entre 'A' et 'F'.
    */
   static char toDigit(int n) {
      if (0 <= n && n <= 9) {
         return (char) (n + (int) '0') ; 
      } else {
         return (char) (n - 10 + (int) 'A') ; 
      }
   }

   /**
    * Convertit un caractre en tableau de caractres unicode.
    * Exemple : charToUnicode(' ') == {'0', '0', '2', '0'}
    */
   static char[] charToUnicode(char c) {
      int val = (int) c ; 
      char[] tab = new char[4] ;
      for (int i = 3 ; 0 <= i ; i--) {
         tab[i] = toDigit(val % 16) ;
         val = val / 16 ;
      }
      return tab ; 
   }

   /**
    * Convertit le caractre c en une chaine de caractres
    * qui le reprsente.
    */
   static String charToString(char c) {
      if (c == '\'') {
         return "'\\\''" ; 
      } else if (c == 92) {
         return "'\\\\'" ; 
      } else if (32 <= c && c <= 127) {
         return "'" + c + "'" ; 
      } else if (c == 10) {
         return "'\\n'" ;           // Il y avait un bug ici
      } else if (c == 13) {
         return "'\\r'" ;           // Il y avait un bug ici aussi
      } else {
         String s = new String(charToUnicode(c)) ;
         return "'\\u" + s + "'" ; 
      }
   }

   /**
    * Convertit une chaine s en une chaine de caractres qui la 
    * reprsente.
    * Exemple : s = 012\345"67'89 --> "012\\345\"67'89"
    */
   static String stringToString(String s) {
      int length = 0 ; 
      for (int i = 0 ; i < s.length() ; i++) {
         char c = s.charAt(i) ; 
         if (c == 10 || c == 13 || c == '\\' || c == '\"') {
            length = length + 2 ; 
         } else if (c < 32 || c > 127) {
            length = length + 6 ; 
         } else {
            length = length + 1 ; 
         }
      }
      length = length + 2 ; 
      char[] t = new char[length] ; 
      t[0] = '"' ; 
      int j = 1 ; 
      for (int i = 0 ; i < s.length() ; i++) {
         char c = s.charAt(i) ; 
         if (c == '\\' || c == '\"') {
            t[j] = '\\' ; 
            t[j + 1] = c ; 
            j = j + 2 ; 
         } else if (c == 10) {
            t[j] = '\\' ;
            t[j + 1] = 'n' ; 
            j = j + 2 ; 
         } else if (c == 13) {
            t[j] = '\\' ;
            t[j + 1] = 'r' ; 
            j = j + 2 ; 
         } else if (c < 32 || c > 127) {
            char[] tab = charToUnicode(c) ; 
            t[j] = '\\' ; 
            t[j + 1] = 'u' ; 
            t[j + 2] = tab[0] ; 
            t[j + 3] = tab[1] ; 
            t[j + 4] = tab[2] ; 
            t[j + 5] = tab[3] ; 
            j = j + 6 ; 
         } else {
            t[j] = c ; 
            j = j + 1 ; 
         }
      }
      t[j] = '"' ; 
      return new String(t) ; 
   }


   /**
    * Evalue cette valeur.
    */
   public Object eval() {
      return valeur ; 
   }


   /**
    * Conversion en chaine de caractres.
    */
   public String toString() {
      return nom ; 
   }


}
